#include <stdio.h>
#include <stdint.h>
#include <math.h>

#include <omp.h>

double inner(const double *p, const double *q, const int k)
{
    int d;
    double product=0;
    for (d = 0; d < k; d++)
        product += p[d]*q[d];
    return product;
}

void axpy(double *p, const double *q, const int k, const double a)
{
    int d;
    for (d = 0; d < k; d++)
        p[d] += a*q[d];
}

void cz(const double *p, const double *q, const double *w, const double *h, const int k,
        const int32_t *idx, const int32_t *nz_j,  const size_t m, const size_t start, const double omega,
        const double *y_sigma, double *ZQ) {

#pragma omp parallel for schedule(dynamic)
    for (size_t i = 0; i < m; i++) {
        const double *pi = p+(i+start)*k, *wi = w+(i+start)*k;
        double *zq_i = ZQ + i*k;
        for (size_t l = idx[i]; l < idx[i+1]; l++) {
            const size_t j = nz_j[l];
            double zij = inner(pi, h+j*k, k) + inner(wi, q+j*k, k);
            zij *= y_sigma[l]*(1-y_sigma[l])-omega;
            axpy(zq_i, q+j*k, k, zij);
        }
    }
}

double cpred_val_sogram(const double *p, const double *q, const int k,
        const double *pt, const double *qt, const int k1,
        const size_t nnz,
        const int32_t *is, const int32_t *js, const double omega,
        const double *y_label, double *bp, double*bq) {

    double p_loss = 0;
#pragma omp parallel for schedule(dynamic) reduction(+:p_loss)
    for (size_t l = 0; l < nnz; l++) {
        const double *pl = p+l*k;
        const double *ql = q+l*k;

        const double *ptl = pt+is[l]*k1;
        const double *qtl = qt+js[l]*k1;

        const double y_hat = inner(pl, ql, k);
        const double y_imp = inner(ptl, qtl, k1);

        double diff = y_hat - y_imp;

        double residual = omega/2*diff*diff;

        double yz = y_label[l]*y_hat;
        double expyy, y_sigma;
        if (yz > 0){
            expyy = exp(-yz);
            y_sigma = expyy/(1+expyy);
            p_loss += log(1+expyy)-residual;
        }
        else {
            expyy = exp(yz);
            y_sigma= 1/(1+expyy);
            p_loss += -yz+log(1+expyy)-residual;
        }
        double bl = -y_label[l]*y_sigma - omega*diff;

        for (int d = 0; d < k; d++) {
            bp[l*k+d]  = bl*pl[d];
            bq[l*k+d]  = bl*ql[d];
        }
    }
    return p_loss;
}

double cpred_val(const double *p, const double *q, const int k,
        const int32_t *idx, const int32_t *nz_j,  const size_t m,
        const double *R, const double *S, const double omega,
        const double *y_label, const double *y_imp, double *y_hat, double*y_sigma) {

    double p_loss = 0;
#pragma omp parallel for schedule(dynamic) reduction(+:p_loss)
    for (size_t i = 0; i < m; i++) {
        const double *pi = p+i*k;
        for (size_t l = idx[i]; l < idx[i+1]; l++) {
            const size_t j = nz_j[l];

            y_hat[l] = inner(pi, q+j*k, k);
            double diff = y_hat[l] - y_imp[l];

            const double Ri = R[i];
            const double Sj = S[j];
            double residual = omega/2*Ri*Sj*diff*diff;

            double yz = y_label[l]*y_hat[l];
            double expyy;
            if (yz > 0){
                expyy = exp(-yz);
                y_sigma[l] = expyy/(1+expyy) + 1e-9;
                p_loss += log(1+expyy)-residual;
            }
            else {
                expyy = exp(yz);
                y_sigma[l] = 1/(1+expyy) + 1e-9;
                p_loss += -yz+log(1+expyy)-residual;
            }
        }
    }
    return p_loss;
}

void cBQ(const double *q, const int k, const int32_t *idx, const int32_t *nz_j, 
        const size_t m, const size_t start, const double *R, const double *S,
        const double omega, const double *y_label, const double *y_imp, const double *y_hat,
        const double *y_sigma, double *BQ) {

#pragma omp parallel for schedule(dynamic)
    for (size_t i = 0; i < m; i++) {
        double *bq_i = BQ + i*k;
        for (size_t l = idx[i]; l < idx[i+1]; l++) {
            const size_t j = nz_j[l];

            const double Ri = R[start+i];
            const double Sj = S[j];

            double diff = y_hat[l] - y_imp[l];
            double bij = -y_label[l]*y_sigma[l] - omega*Ri*Sj*diff;
            axpy(bq_i, q+j*k, k, bij);
        }
    }
}
