import os
import math
import time

import numpy as np
from scipy.sparse import csr_matrix

import ctypes
from numpy.ctypeslib import ndpointer

import tensorflow.compat.v1 as tf
from tensorflow.compat.v1 import SparseTensorValue as STV

from utilities import predict, vectorize, inverse_vectorize, calc_norm, zero_tensor
from utilities import Y as Yclass
import pdb

def Rop(f, weights, v):
    """Implementation of R operator
    Args:
        f: any function of weights
        weights: list of tensors.
        v: vector for right multiplication
    Returns:
        Jv: Jaccobian vector product, length same as
            the number of output of f
    """
    if type(f) == list:
        u = [tf.zeros_like(ff) for ff in f]
    else:
        u = tf.zeros_like(f)  # dummy variable
    g = tf.gradients(ys=f, xs=weights, grad_ys=u)
    return tf.gradients(ys=g, xs=u, grad_ys=v)


class newton_cg(object):
    def __init__(self, config, sess, network):
        """
        initialize operations and vairables that will be used in newton
        args:
            sess: tensorflow session
            outputs: output of the neural network (pre-softmax layer)
            loss: function to calculate loss
       """
        super(newton_cg, self).__init__()
        self.network = network
        self.u, self.v, self.p, self.q = network
        self.sess = sess

        lib = ctypes.cdll.LoadLibrary("./c_funs/cfuns.so")

        self.cz = lib.cz
        self.cz.restype = None
        self.cz.argtypes = [
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ctypes.c_int,
                ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
                ctypes.c_size_t,
                ctypes.c_size_t,
                ctypes.c_double,
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ]

        self.cpred_val = lib.cpred_val
        self.cpred_val.restype = ctypes.c_double
        self.cpred_val.argtypes = [
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ctypes.c_int,
                ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
                ctypes.c_size_t,
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ctypes.c_double,
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ]

        self.cBQ = lib.cBQ
        self.cBQ.restype = None
        self.cBQ.argtypes = [
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ctypes.c_int,
                ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
                ctypes.c_size_t,
                ctypes.c_size_t,
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ctypes.c_double,
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ]

        self.config = config

        #TODO define scope name
        self.param = tf.trainable_variables()

        self.param_u = tf.trainable_variables(scope="u_model")
        self.param_v = tf.trainable_variables(scope="v_model")

        FLOAT = tf.float64
        self.FLOAT = FLOAT
        self.NP_FLOAT = np.float64

        self.CGiter = 0

        model_weight_u = vectorize(self.param_u)
        model_weight_v = vectorize(self.param_v)

        self.p_loss = 0
        self.p_loss_holder = tf.placeholder(FLOAT)
        self.f = tf.Variable(0., dtype=FLOAT, trainable=False)
        self.nloss = tf.Variable(0., dtype=FLOAT, trainable=False)
        self.reg = tf.Variable(0., dtype=FLOAT, trainable=False)

        self.adagrad_init_accum_value = 0.1
        self.adagrad_eps = 1e-7

        # initial variables cached in CPU
        self.P = np.zeros([self.config.m, self.config.k], dtype=self.NP_FLOAT)
        self.Q = np.zeros([self.config.n, self.config.k], dtype=self.NP_FLOAT)

        self.Pt = np.ones([self.config.m, self.config.k1], dtype=self.NP_FLOAT)*self.config.r*math.sqrt(1/self.config.k1)
        self.Qt = np.ones([self.config.n, self.config.k1], dtype=self.NP_FLOAT)*math.sqrt(1/self.config.k1)

        # todo: try tf.linalg.tensor_diag
        self.R = np.ones([self.config.m], dtype=self.NP_FLOAT)
        self.S = np.ones([self.config.n], dtype=self.NP_FLOAT)

        # calculate tPc, tQc
        RP = self.R[:, np.newaxis]*self.Pt
        SQ = self.S[:, np.newaxis]*self.Qt
        tPc = np.zeros([self.config.k1, self.config.k1], dtype=self.NP_FLOAT)
        tQc = np.zeros([self.config.k1, self.config.k1], dtype=self.NP_FLOAT)
        num_segment = math.ceil(self.config.m/self.config.bsize)
        for i in range(num_segment):
            idx = np.arange(i * self.config.bsize, min((i+1) * self.config.bsize, self.config.m))
            tPc += RP[idx].T.dot(RP[idx])
        num_segment = math.ceil(self.config.n/self.config.bsize)
        for i in range(num_segment):
            idx = np.arange(i * self.config.bsize, min((i+1) * self.config.bsize, self.config.n))
            tQc += SQ[idx].T.dot(SQ[idx])
        self.tPC_tQC_F = (tPc * tQc).sum()

        self.R_holder = tf.placeholder(dtype=self.FLOAT)
        self.S_holder = tf.placeholder(dtype=self.FLOAT)

        self.W = np.zeros([self.config.m, self.config.k], dtype=self.NP_FLOAT)
        self.H = np.zeros([self.config.n, self.config.k], dtype=self.NP_FLOAT)

        # initial Grammian variabels in GPU
        kk_shape = tf.zeros(shape=[self.config.k, self.config.k], dtype=FLOAT)
        self.Pc = tf.Variable(kk_shape, dtype=FLOAT, trainable=False)
        self.Qc = tf.Variable(kk_shape, dtype=FLOAT, trainable=False)

        self.Wc = tf.Variable(kk_shape, dtype=FLOAT, trainable=False)
        self.Hc = tf.Variable(kk_shape, dtype=FLOAT, trainable=False)

        k1k_shape = tf.zeros(shape=[self.config.k1, self.config.k], dtype=FLOAT)
        self.hPc = tf.Variable(k1k_shape, dtype=FLOAT, trainable=False)
        self.hQc = tf.Variable(k1k_shape, dtype=FLOAT, trainable=False)

        zeros_u = tf.zeros(model_weight_u.get_shape(), dtype=FLOAT)
        zeros_v = tf.zeros(model_weight_v.get_shape(), dtype=FLOAT)

        # initial variable used in gradient
        self.g_u = tf.Variable(zeros_u, dtype=FLOAT, trainable=False)
        self.g_v = tf.Variable(zeros_v, dtype=FLOAT, trainable=False)

        self.ada_u = tf.Variable(zeros_u+self.adagrad_init_accum_value, dtype=FLOAT, trainable=False)
        self.ada_v = tf.Variable(zeros_v+self.adagrad_init_accum_value, dtype=FLOAT, trainable=False)

        # initial variable used in CG
        self.r_u = tf.Variable(zeros_u, dtype=FLOAT, trainable=False)
        self.r_v = tf.Variable(zeros_v, dtype=FLOAT, trainable=False)

        self.d_u = tf.Variable(zeros_u, dtype=FLOAT, trainable=False)
        self.d_v = tf.Variable(zeros_v, dtype=FLOAT, trainable=False)

        self.Gd_u = tf.Variable(zeros_u, dtype=FLOAT, trainable=False)
        self.Gd_v = tf.Variable(zeros_v, dtype=FLOAT, trainable=False)

        self.s_u = tf.Variable(zeros_u, dtype=FLOAT, trainable=False)
        self.s_v = tf.Variable(zeros_v, dtype=FLOAT, trainable=False)


        # rTr, cgtol and beta to be used in CG
        self.rTr = tf.Variable(0., dtype=FLOAT, trainable=False)
        self.cgtol = tf.Variable(0., dtype=FLOAT, trainable=False)
        self.beta = tf.Variable(0., dtype=FLOAT, trainable=False)

        # placeholder alpha, old_alpha and lambda
        self.alpha = tf.placeholder(FLOAT, shape=[])
        self.old_alpha = tf.placeholder(FLOAT, shape=[])
        self._lambda = tf.placeholder(FLOAT, shape=[])

        self.gnorm = calc_norm(tf.concat([self.g_u, self.g_v], axis=0))

        self.CG, self.update_v = self._CG()
        self.init_cg_vars = self._init_cg_vars()
        self.update_gs = tf.tensordot(self.s_u, self.g_u, axes=1) + tf.tensordot(self.s_v, self.g_v, axes=1)
        self.update_sGs = 0.5*(tf.tensordot(self.s_u, -self.g_u -self.r_u -self._lambda*self.s_u, axes=1) +
                               tf.tensordot(self.s_v, -self.g_v -self.r_v -self._lambda*self.s_v, axes=1))
        self.update_model = self._update_model()

        self.cal_nloss = self._cal_nloss()
        self.cal_reg = self._cal_reg()
        self.cal_fun_val = self._cal_fun_val()

        self.tp = tf.placeholder(dtype=self.FLOAT)
        self.tq = tf.placeholder(dtype=self.FLOAT)

        self.zero_Pc_hPc = self._zero_gramian(self.Pc, self.hPc)
        self.zero_Qc_hQc = self._zero_gramian(self.Qc, self.hQc)

        self.cal_gramian_P = self._cal_gramian(self.p, self.tp, self.Pc, self.hPc, self.R_holder)
        self.cal_gramian_Q = self._cal_gramian(self.q, self.tq, self.Qc, self.hQc, self.S_holder)
        self.holder_bq = tf.placeholder(dtype=self.FLOAT)
        self.holder_bp = tf.placeholder(dtype=self.FLOAT)

        self.holder_p = tf.placeholder(self.FLOAT)
        self.holder_q = tf.placeholder(self.FLOAT)

        self.zero_g_u = zero_tensor(self.g_u)
        self.zero_g_v = zero_tensor(self.g_v)

        self.batch_grad_u = self._cal_batch_grad(self.holder_bq, self.holder_p, self.tp,
                self.Qc, self.hQc, self.p, self.param_u, self.g_u, self.R_holder)
        self.batch_grad_v = self._cal_batch_grad(self.holder_bp, self.holder_q, self.tq,
                self.Pc, self.hPc, self.q, self.param_v, self.g_v, self.S_holder)

        self.reg_grad_u = self._add_reg_grad(self.param_u, self.g_u)
        self.reg_grad_v = self._add_reg_grad(self.param_v, self.g_v)

        self.holder_w = tf.placeholder(self.FLOAT)
        self.holder_h = tf.placeholder(self.FLOAT)

        self.zero_Wc = zero_tensor(self.Wc)
        self.zero_Hc = zero_tensor(self.Hc)

        self.cal_lossJd_u = self._cal_lossJd(self.d_u, self.param_u, self.p)
        self.cal_Wc = self._cal_Wc(self.Wc, self.holder_p, self.cal_lossJd_u, self.R_holder)

        self.cal_lossJd_v = self._cal_lossJd(self.d_v, self.param_v, self.q)
        self.cal_Hc = self._cal_Wc(self.Hc, self.holder_q, self.cal_lossJd_v, self.S_holder)

        self.holder_zq = tf.placeholder(dtype=self.FLOAT)
        self.holder_zp = tf.placeholder(dtype=self.FLOAT)

        self.batch_Gd_u = self._cal_lossGd(self.holder_zq, self.holder_w, self.holder_p,
                self.Qc, self.Hc, self.p, self.param_u, self.Gd_u, self.R_holder)

        self.batch_Gd_v = self._cal_lossGd(self.holder_zp, self.holder_h, self.holder_q,
                self.Pc, self.Wc, self.q, self.param_v, self.Gd_v, self.S_holder)

        self.zero_Gd_u = zero_tensor(self.Gd_u)
        self.zero_Gd_v = zero_tensor(self.Gd_v)

        self.reg_Gd_u = self._add_reg_Gd(self.Gd_u, self.d_u)
        self.reg_Gd_v = self._add_reg_Gd(self.Gd_v, self.d_v)

        self.update_adagrad = tf.group(*[
            tf.assign(self.ada_u, self.ada_u + self.g_u*self.g_u),
            tf.assign(self.ada_v, self.ada_v + self.g_v*self.g_v)
        ])

        self.adagrad_direction = tf.group(*[
            tf.assign(self.s_u, -self.g_u*tf.math.rsqrt(self.ada_u+self.adagrad_eps)),
            tf.assign(self.s_v, -self.g_v*tf.math.rsqrt(self.ada_v+self.adagrad_eps))
        ])

        self.gradient_direction = tf.group(*[tf.assign(self.s_u, -self.g_u), tf.assign(self.s_v, -self.g_v)])


    def profile(self, *args, **kwargs):
        if self.config.profile:
            print(*args, **kwargs)


    def cal_pred_val(self, Y, P, Q, R, S):
        """
        Calculate L^+ part's LogLoss via expyy values of nonzeros in Y
        """
        indptr = Y.Label.indptr
        indices = Y.Label.indices
        m = Y.Label.shape[0]
        omega = self.config.omega

        self.p_loss = self.cpred_val(P, Q, self.config.k, indptr, indices, m, R, S, omega, 
                Y.Label.data, Y.Y_imp.data, Y.Y_hat.data, Y.Sigma.data)


    def cal_ival(self, Y):
        coos = Y.Y_imp.nonzero()
        ps = self.Pt[coos[0]]
        qs = self.Qt[coos[1]]
        Y.Y_imp.data = (ps*qs).sum(1)

    def transpose(self, Y):
        YT = Yclass(Y.n, Y.m, Y.nnzs)

        nonzero = Y.Label.transpose().nonzero()
        YT.Label = csr_matrix((np.array(Y.Label.transpose()[nonzero]).squeeze(), nonzero), shape=(Y.n, Y.m))
        
        YT.Sigma = csr_matrix((np.array(Y.Sigma.transpose()[nonzero]).squeeze(), nonzero), shape=(Y.n, Y.m))

        YT.Y_hat = csr_matrix((np.array(Y.Y_hat.transpose()[nonzero]).squeeze(), nonzero), shape=(Y.n, Y.m))
        YT.Y_imp = csr_matrix((np.array(Y.Y_imp.transpose()[nonzero]).squeeze(), nonzero), shape=(Y.n, Y.m))
        
        return YT

    def _cal_gramian(self, p, tp, Pc, hPc, R_holder):
        RP = tf.multiply(R_holder[:, tf.newaxis], p)
        update_Pc = tf.assign(Pc, Pc + tf.transpose(p) @ RP)
        update_hPc = tf.assign(hPc, hPc + tf.transpose(tp) @ RP)
        return tf.group(*[update_Pc, update_hPc])
    
    def _zero_gramian(self, Pc, hPc):
        zero_Pc = zero_tensor(Pc)
        zero_hPc = zero_tensor(hPc)
        return tf.group(*[zero_Pc, zero_hPc])

    def _cal_nloss(self):
        PcQc_F = tf.reduce_sum(tf.math.multiply(self.Pc, self.Qc), [0,1])
        hPc_hQc_F = tf.reduce_sum(tf.math.multiply(self.hPc, self.hQc), [0,1])
        # todo: add tPc_tQc_F to nloss?
        return tf.assign(self.nloss, self.tPC_tQC_F/2 + PcQc_F/2 - hPc_hQc_F)

    def _cal_reg(self):
        model_weight = vectorize(self.param)
        reg = (calc_norm(model_weight))**2
        return tf.assign(self.reg, reg*self.config._lambda/2)

    def _cal_lossJd(self, d_u, param_u, p):
        d = inverse_vectorize(d_u, param_u)
        Jd = Rop(p, param_u, d)
        return Jd[0]

    def _cal_Wc(self, Wc, holder_p, Jd_u, R_holder):
        RP = tf.multiply(R_holder[:, tf.newaxis], holder_p)
        update_Wc = tf.assign(Wc, Wc + tf.transpose(Jd_u) @ holder_p)
        return update_Wc


    def _cal_lossGd(self, holder_zq, holder_w,
                    holder_p, Qc, Hc, p, param_u, Gd, R_holder):
        pre_Gd_neg = tf.multiply(R_holder[:, tf.newaxis], holder_w @ Qc + holder_p @ Hc)
        pre_Gd = holder_zq + self.config.omega*pre_Gd_neg
        Gd_u = tf.gradients(ys=p, xs=param_u, grad_ys=pre_Gd)
        Gd_u = vectorize(Gd_u)
        update_Gd = tf.assign(Gd, Gd+Gd_u)
        return update_Gd


    def microbatches(self, input):
        m = input.shape[0]
        num_segment = math.ceil(m/self.config.bsize)
        for i in range(num_segment):
            load_time = time.time()
            idx = np.arange(i * self.config.bsize, min((i+1) * self.config.bsize, m))
            start, end = idx[0], idx[-1]+1
            subset = slice(start, end)
            batch_input = input[subset]
            batch_input = STV(np.array(batch_input.nonzero()).transpose(), batch_input.data, batch_input.shape)
            self.config.elapsed_time += time.time() - load_time
            
            yield subset, batch_input


    def cal_output(self, zero_gramian, cal_gramian, input, place_holder_u, tp, p, tP, P, R_holder, R):
        """
        A function to evaluate either function value, global gradient or sub-sampled Gv
        """

        self.sess.run(zero_gramian)

        u = place_holder_u
        for subset, batch_input in self.microbatches(input):
            P[subset],_ = self.sess.run([p, cal_gramian], feed_dict={
                u: batch_input, tp: tP[subset], R_holder: R[subset]})

    def _cal_batch_grad(self, holder_bq, holder_p, holder_tp, Qc, hQc, p, param, g, R_holder):
        pre_grad_neg = tf.multiply(R_holder[:, tf.newaxis], holder_p @ Qc - holder_tp @ hQc)
        pre_grad = holder_bq + self.config.omega*pre_grad_neg
        grad = tf.gradients(ys=p, xs=param, grad_ys=pre_grad)
        grad = vectorize(grad)
        update_grad = tf.assign(g, g+grad)
        return update_grad

    def _add_reg_grad(self, param, g):
        model_weight = vectorize(param)
        reg_grad = model_weight*self.config._lambda
        return tf.assign(g, g+reg_grad)

    def _add_reg_Gd(self, Gd_u, d_u):
        return tf.assign(Gd_u, Gd_u+self.config._lambda*d_u)
        
    def cal_BQ(self, Y, Q, subset, R, S):
        start = subset.start

        label_ = Y.Label[subset]
        indptr = label_.indptr
        indices = label_.indices

        m_sub = label_.shape[0]

        omega = self.config.omega

        label_sub = label_.data
        y_imp_sub = Y.Y_imp[subset].data
        y_hat_sub = Y.Y_hat[subset].data
        sigma_sub = Y.Sigma[subset].data

        BQ = np.zeros((m_sub, self.config.k), dtype = self.NP_FLOAT)
        self.cBQ(Q, self.config.k, indptr, indices, m_sub, start, R, S, omega, 
                label_sub, y_imp_sub, y_hat_sub, sigma_sub, BQ)
        return BQ
        

    def cal_grad(self, zero_g, batch_grad, reg_grad, input, Y, P, Q, tP, place_holder_u, p, holder_bq, holder_p, holder_tp, R_holder, R, S):
        # before estimation starts, need to zero out grad
        self.sess.run(zero_g)

        u = place_holder_u
        for subset, batch_input in self.microbatches(input):
            bP = np.ascontiguousarray(P[subset])
            btP = np.ascontiguousarray(tP[subset])
            s = time.time()

            bBQ =  self.cal_BQ(Y, Q, subset, R, S)
            self.profile("bBQ:",time.time()-s)
            s = time.time()
            self.sess.run(batch_grad, feed_dict={u: batch_input, holder_bq: bBQ,
                                                 holder_p: bP, holder_tp: btP,
                                                 R_holder: R[subset]})
            self.profile("gpu_grad:",time.time()-s)
        
        self.sess.run(reg_grad)


    def cal_Jd(self, zero_Wc, cal_lossJd, cal_Wc, input, P, W, place_holder_u, holder_p, R_holder, R):
        W.fill(0)
        self.sess.run(zero_Wc)

        u = place_holder_u
        for subset, batch_input in self.microbatches(input):
            bP = np.ascontiguousarray(P[subset])
            W[subset],_ = self.sess.run([cal_lossJd, cal_Wc],
                                     feed_dict={u: batch_input, holder_p:bP, R_holder: R[subset]})

    def cal_ZQ(self, Y, subset, Q, W, P, H):
        start = subset.start

        label_ = Y.Label[subset]
        indptr = label_.indptr
        indices = label_.indices

        m_sub = label_.shape[0]

        omega = self.config.omega
        sigma_sub = Y.Sigma[subset].data

        ZQ = np.zeros((m_sub, self.config.k), dtype = self.NP_FLOAT)
        self.cz(P, Q, W, H, self.config.k, indptr, indices, m_sub, start, omega, sigma_sub, ZQ)

        return ZQ

    def cal_Gd(self, zero_Gd_u, batch_Gd, reg_Gd_u,  input, Y, P, Q, W, H, place_holder_u, holder_p, holder_w, holder_zq, R_holder, R, S):
        self.sess.run(zero_Gd_u)
        
        u = place_holder_u
        for subset, batch_input in self.microbatches(input):
            bP = np.ascontiguousarray(P[subset])
            bW = np.ascontiguousarray(W[subset])
            s = time.time()

            bZQ = self.cal_ZQ(Y, subset, Q, W, P, H)
            self.profile("bZQ:",time.time()-s)
            
            s = time.time()
            self.sess.run(batch_Gd, feed_dict={u: batch_input, holder_zq: bZQ, holder_p: bP, holder_w: bW, R_holder: R[subset]})
            self.profile("gpu_Gd:",time.time()-s)
        
        self.sess.run(reg_Gd_u)



    def _update_model(self):
        update_model_ops = []

        x = inverse_vectorize(self.s_u, self.param_u)
        for i, p in enumerate(self.param_u):
            op = tf.assign(p, p + (self.alpha-self.old_alpha) * x[i])
            update_model_ops.append(op)

        x = inverse_vectorize(self.s_v, self.param_v)
        for i, p in enumerate(self.param_v):
            op = tf.assign(p, p + (self.alpha-self.old_alpha) * x[i])
            update_model_ops.append(op)
        return tf.group(*update_model_ops)

    def _init_cg_vars(self):
        init_ops = []

        init_r_u = tf.assign(self.r_u, -self.g_u)
        init_r_v = tf.assign(self.r_v, -self.g_v)

        init_d_u = tf.assign(self.d_u, -self.g_u)
        init_d_v = tf.assign(self.d_v, -self.g_v)

        init_s_u = tf.assign(self.s_u, tf.zeros_like(self.g_u))
        init_s_v = tf.assign(self.s_v, tf.zeros_like(self.g_v))

        gnorm = self.gnorm

        init_rTr = tf.assign(self.rTr, gnorm**2)

        init_cgtol = tf.assign(self.cgtol, self.config.xi*gnorm)

        init_ops = [init_r_u, init_r_v, init_d_u,
                    init_d_v, init_s_u, init_s_v, init_rTr, init_cgtol]

        return tf.group(*init_ops)

    def _CG(self):
        """
        CG:
            define operations that will be used in method newton

        Same as the previous loss calculation,
        Gv has been summed over batches when samples were fed into Neural Network.
        """

        def CG_ops():

            dGd = tf.tensordot(self.d_u, self.Gd_u, axes=1) + tf.tensordot(self.d_v, self.Gd_v, axes=1)

            alpha = self.rTr / dGd
            with tf.control_dependencies([alpha]):
                #TODO concat d_u, d_v, s, Gdu Gdv

                update_s_u = tf.assign(self.s_u, self.s_u + alpha * self.d_u, name='update_su_ops')
                update_s_v = tf.assign(self.s_v, self.s_v + alpha * self.d_v, name='update_sv_ops')

                update_r_u = tf.assign(self.r_u, self.r_u - alpha * self.Gd_u, name='update_ru_ops')
                update_r_v = tf.assign(self.r_v, self.r_v - alpha * self.Gd_v, name='update_rv_ops')

                with tf.control_dependencies([update_s_u, update_s_v, update_r_u, update_r_v]):
                    rnewTrnew = calc_norm(tf.concat([update_r_u, update_r_v], axis=0))**2
                    update_beta = tf.assign(self.beta, rnewTrnew / self.rTr)
                    with tf.control_dependencies([update_beta]):
                        update_rTr = tf.assign(self.rTr, rnewTrnew, name='update_rTr_ops')

            return tf.group(*[update_s_u, update_s_v, update_beta, update_rTr])

        def update_v():
            update_d_u = tf.assign(self.d_u, self.r_u + self.beta*self.d_u, name='update_du')
            update_d_v = tf.assign(self.d_v, self.r_v + self.beta*self.d_v, name='update_dv')
            return tf.group(*[update_d_u, update_d_v])

        return (CG_ops(), update_v())

    def _cal_fun_val(self):
        return tf.assign(self.f, self.config.omega*self.cal_nloss + self.p_loss_holder + self.cal_reg)


    def fun_val(self, Utr, V, Y, YT, network):
        u,v,p,q = network
        s = time.time()
        self.cal_output(self.zero_Pc_hPc, self.cal_gramian_P,  Utr, u, self.tp, p, self.Pt, self.P, self.R_holder, self.R)
        self.cal_output(self.zero_Qc_hQc, self.cal_gramian_Q, V, v, self.tq, q, self.Qt, self.Q, self.S_holder, self.S)
        self.profile("output:",time.time()-s)
        
        s = time.time()
        self.cal_pred_val(Y, self.P, self.Q, self.R, self.S)
        self.cal_pred_val(YT, self.Q, self.P, self.S, self.R)
        
        self.profile("ploss:",time.time()-s)
        s = time.time()

        self.sess.run(self.cal_fun_val, feed_dict={self.p_loss_holder:self.p_loss})
        self.profile("nloss:",time.time()-s)

    def run_line_search(self, f_old, Utr, V, Y, YT):
        gs, sGs = self.sess.run([self.update_gs, self.update_sGs], feed_dict={
            self._lambda: self.config._lambda
        })

        if self.config.alpha == 0:
            alpha = self.alpha_prev if self.config.from_prev else 1
            max_linesearch = 50
        else:
            alpha = self.config.alpha
            max_linesearch = 1

        for line_search in range(1, max_linesearch + 1):
            old_alpha = 0 if line_search == 1 else alpha/0.5

            self.sess.run(self.update_model, feed_dict={
                self.alpha:alpha, self.old_alpha:old_alpha
                })

            prered = alpha*gs + (alpha**2)*sGs

            self.fun_val(Utr, V, Y, YT, self.network)
            f = self.sess.run(self.f)

            actred = f - f_old

            line_search_stop_cond = actred <= self.config.eta*alpha*gs
            if line_search_stop_cond or max_linesearch == 1:
                break

            alpha *= 0.5

        return line_search, alpha, actred, prered, f

    def gauss_newton_direction(self, Utr, V, Y, YT):
        self.sess.run(self.init_cg_vars)
        cgtol = self.sess.run(self.cgtol)

        avg_cg_time = 0.0

        for CGiter in range(1, self.config.CGmax+1):

            cg_time = time.time()

            s = time.time()
            self.cal_Jd(self.zero_Wc, self.cal_lossJd_u,
                    self.cal_Wc, Utr, self.P, self.W, self.u, self.holder_p, self.R_holder, self.R)

            self.cal_Jd(self.zero_Hc, self.cal_lossJd_v,
                    self.cal_Hc, V, self.Q, self.H, self.v, self.holder_q, self.S_holder, self.S)
            self.profile("Jd:",time.time()-s)

            s = time.time()
            self.cal_Gd(self.zero_Gd_u, self.batch_Gd_u, self.reg_Gd_u, Utr, Y, self.P, self.Q,
                    self.W, self.H, self.u, self.holder_p, self.holder_w, self.holder_zq, self.R_holder, self.R, self.S)

            self.cal_Gd(self.zero_Gd_v, self.batch_Gd_v, self.reg_Gd_v, V, YT, self.Q, self.P,
                    self.H, self.W, self.v, self.holder_q, self.holder_h, self.holder_zp, self.S_holder, self.S, self.R)
            self.profile("Gd:",time.time()-s)

            avg_cg_time += time.time() - cg_time

            self.sess.run(self.CG)

            rnewTrnew = self.sess.run(self.rTr)

            if rnewTrnew**0.5 <= cgtol or CGiter == self.config.CGmax:
                break

            self.sess.run(self.update_v)

        #print('Avg time per Gv iteration: {:.5f} s\r\n'.format(avg_cg_time/CGiter))
        return CGiter

    def newton(self, Y, Utr, V, val_batch, saver):
        """
        Conduct newton steps for training
        args:
            full_batch & val_batch: provide training set and validation set. The function will
                save the best model evaluted on validation set for future prediction.
            network: a tuple contains (x, y, loss, outputs).
            test_network: a tuple similar to argument network. If you use layers which behave differently
                in tesse such as batchnorm, a separate test_network is needed.
        return:
            None
        """
        # check whether data is valid
        assert (Y.m == Utr.shape[0]) and (Y.n == V.shape[0])


        tf.summary.scalar('loss', self.f)
        merged = tf.summary.merge_all()
        train_writer = tf.summary.FileWriter('./summary/train', self.sess.graph)

        print(self.config.args)
        if not self.config.screen_log_only:
            log_file = open(self.config.log_file, 'w')
            print(self.config.args, file=log_file)
        
        self.cal_ival(Y)

        YT = self.transpose(Y)

        u, v, p, q = self.network
        self.fun_val(Utr, V, Y, YT, self.network)
        f = self.sess.run(self.f)

        best_acc = 0.0


        total_running_time = 0.0
        self.config.elapsed_time = 0.0
        total_CG = 0
        self.alpha_prev = 1
        for k in range(self.config.iter_max):

            # randomly select the batch for Gv estimation
            #TODO add subsample-Gauss
            """
            idx = np.random.choice(np.arange(0, full_labels.shape[0]),
                    size=self.config.GNsize, replace=False)

            mini_inputs = full_inputs[idx]
            mini_labels = full_labels[idx]
            """

            start = time.time()

            s = time.time()
            self.cal_grad(self.zero_g_u, self.batch_grad_u, self.reg_grad_u,
                Utr, Y, self.P, self.Q, self.Pt, u, p, self.holder_bq, self.holder_p, self.tp, self.R_holder, self.R, self.S)

            self.cal_grad(self.zero_g_v, self.batch_grad_v, self.reg_grad_v,
                V, YT, self.Q, self.P, self.Qt, v, q, self.holder_bp, self.holder_q, self.tq, self.S_holder, self.S, self.R)
            self.profile("grad:",time.time()-s)

            gnorm = self.sess.run(self.gnorm)

            if self.config.CGmax == 0:
                if self.config.adagrad:
                    self.sess.run(self.update_adagrad)
                    self.sess.run(self.adagrad_direction)
                else:
                    self.sess.run(self.gradient_direction)
                CGiter = 0
            else:
                CGiter = self.gauss_newton_direction(Utr, V, Y, YT)

            line_search, alpha, actred, prered, f = self.run_line_search(f, Utr, V, Y, YT)
            self.alpha_prev = alpha
            if actred / prered > self.config.thres:
                self.alpha_prev *= 4

            summary = self.sess.run(merged)
            train_writer.add_summary(summary, k)

            # exclude data loading time for fair comparison
            end = time.time()

            end = end - self.config.elapsed_time
            total_running_time += end-start

            self.config.elapsed_time = 0.0

            total_CG += CGiter

            output_str = 'iter: {} f: {:.3f} |g|: {:.5f} alpha: {:.3e} #line_search: {} #CG: {} ratio: {:.3f} actred: {:.5f} prered: {:.5f} time: {:.3f}'.\
                            format(k, f, gnorm, alpha, line_search, CGiter, actred/prered, actred, prered, total_running_time)
            print(output_str)
            if not self.config.screen_log_only:
                print(output_str, file=log_file)

            if val_batch is not None and (k+1) % 1 == 0:
                # Evaluate the performance after every Newton Step
                Ute, Yte = val_batch
                psprecision, ndcg = predict(self.sess, self.network, Yte.Label, Ute, V, self.config)
                precision = "p@k"+ " " + " ".join([str(round(i*100,3)) for i in psprecision])
                ndcg = "ndcg@k" + " " + " ".join([str(round(i*100,3)) for i in ndcg])
                print(precision + "\t" + ndcg)
                print(precision + "\t" + ndcg, file=log_file)

            log_file.flush()
                
                #TODO Save the best model using validation set
                # if val_acc > best_acc:
                #     best_acc = val_acc
                #     checkpoint_path = self.config.model_file
                #     save_path = saver.save(self.sess, checkpoint_path)
                #     print('Best model saved in {}\r\n'.format(save_path))

        if val_batch is None:
            checkpoint_path = self.config.model_file
            save_path = saver.save(self.sess, checkpoint_path)
            output_str = 'Model at the last iteration saved in {}\r\n'.format(save_path)
            output_str += 'total_#CG {} | total running time {:.3f}s'.format(total_CG, total_running_time)
        else:
            output_str = 'psprecision: {} | ndcg: {} | total_#CG {} | total running time {:.3f}s'.\
                format(psprecision, ndcg, total_CG, total_running_time)
        print(output_str, file=log_file)
        if not self.config.screen_log_only:
            print(output_str, file=log_file)
            log_file.close()
