import matplotlib.pyplot as plt
import re
from pathlib import Path

logs = Path('./exp_logs/')
figs = Path('./figs/')
figs.mkdir(exist_ok=True)

K = 5
vnames = ['epoch', 'f', 'time', 'ls', 'cg', 'map', 'datapass', 'step']
for metric in ['p', 'ndcg']:
    vnames += [metric+str(i+1) for i in range(K)]

xlabel_map = {
    'time': 'Time (s)',
}
ylabel_map = {
    'f': 'Relative difference in objective value',
    'map': 'MAP@5',
}

leg2vname2values = {}

def path2legend_color_marker(path):
    if 'fixed.ss' in path:
        return 'Sampling', 'lime', 'v'
    elif 'mb.ss' in path:
        return 'Sampling-diag', 'green', '^'
    elif 'fixed.sogram' in path:
        return 'SOGram', 'orange', 's'
    elif 'mb.sogram' in path:
        return 'SOGram-diag', 'red', 'D'
    elif '.grad' in path:
        return 'GD', 'cyan', 'X'
    elif 'adagradls' in path:
        return 'GD-diag', 'blue', 'P'
    elif 'gn' in path:
        return 'Newton', 'black', 'o'
    else:
        return None, None, None

def read_file(path, data):
    global leg2vname2values, K, vnames
    leg = path2legend_color_marker(path)[0]
    
    data2obs = {
        'ml1m': 517770,
        'ml10m': 4505820,
        'wiki-simple': 5478549,
    }
    
    leg2vname2values[leg] = {vname: [] for vname in vnames}
    with open(path) as f:
        for line in f:
            line = line.strip()
            if 'nan' in line:
                break
            def collect(pattern):
                match = re.search(pattern, line)
                for vname in vnames:
                    if vname in match.groupdict():
                        leg2vname2values[leg][vname].append(float(match.group(vname)))
                
            fp = lambda name: '(?P<{}>[\d.eE+-]+)'.format(name)
            if line.startswith('iter: '):
                pattern = 'iter: {} f: {} .*time: {}'.format(fp('epoch'), fp('f'), fp('time'))
                collect(pattern)
                
                if '#line_search' in line:
                    pattern = '#line_search: {} #CG: {}'.format(fp('ls'), fp('cg'))
                    collect(pattern)
            elif line.startswith('p@k'):
                pattern_metric = lambda metric: ' '.join(fp(metric+str(i+1)) for i in range(K))
                pattern = 'p@k {}\tndcg@k {}'.format(pattern_metric('p'), pattern_metric('ndcg'))
                collect(pattern)
                avg = lambda l: sum(l) / len(l)
                map_ = avg([leg2vname2values[leg]['p'+str(i+1)][-1] for i in range(K)]) / 100.0
                leg2vname2values[leg]['map'].append(map_)



def plot(data, xlabel, ylabel, ignore=None, xlim=None, ylim=None, suffix=''):
    global leg2vname2values, logs, figs, xlabel_map, ylabel_map

    
    def order(path):
        name = path.as_posix()
        ans = 0
        if 'ss' in name:
            ans -= 32
        elif 'sogram' in name:
            ans -= 16
        if 'fixed' in name:
            ans -= 8
        elif '.grad' in name:
            ans -= 8
        elif 'ada' in name:
            ans -= 4
        
        return ans
    
    paths = [str(path) for path in sorted((logs/data).iterdir(), key=order)]
    paths = [path for path in paths if ignore is None or all(i not in path for i in ignore)]
    
    for path in paths:
        read_file(path, data)

    opt = min(min(leg2vname2values[path2legend_color_marker(path)[0]]['f']) for path in paths)
    
    fig, axes = plt.subplots(1, 1, figsize=(6, 5))
    legends = []
    for path in paths:
        legend, color, marker = path2legend_color_marker(path)
        vname2values = leg2vname2values[legend]
        ax = axes
    
        xs = vname2values[xlabel]
        ys = vname2values[ylabel]
        if len(ys) == 0:
            print(path, ylabel)
            continue
        
        legends.append(legend)
        every = len(xs) // len(ys)
        xs = xs[every-1::every]
        
        if len(xs) > len(ys):
            xs = xs[:len(ys)]
        elif len(ys) > len(xs):
            ys = ys[:len(xs)]

        n = len(ys)
        n_markers = 10
        if xlim is not None and xs[-1] > xlim:
            n_markers *= xs[-1] / xlim
        markevery = max(1, int(n // n_markers)) if n >= 10 else 1
        ls = 'solid' if ('-diag' in legend or 'Newton' in legend) else 'dashed'
        
        ax.plot(xs, ys, c=color, marker=marker, markevery=markevery, ls=ls, markersize=10, linewidth=3)
        ax.set_xlabel(xlabel_map[xlabel], fontsize='x-large')
        ax.set_ylabel(ylabel_map[ylabel], fontsize='x-large')

    if ylim is not None:
        axes.set_ylim(bottom=ylim[0], top=ylim[1])
    if xlim is not None:
        axes.set_xlim(left=-xlim // 20, right=xlim)
    
    loc = 'upper right' if ylabel == 'f' else 'lower right'
    plt.legend(legends, fontsize='x-large', loc=loc)
    plt.title(data, fontsize=25)
    fmt = 'eps'
    plt.savefig((figs/('{}-{}-{}{}.{}'.format(data, xlabel, ylabel, suffix, fmt))).as_posix(), format=fmt)



def plot_f_map(data, ignore, ylim_f, xlim=dict(), suffix=''):
    for xlabel in ['time']:
        for ylabel in ['f', 'map']:
            ylim = (-0.1, ylim_f) if ylabel == 'f' else None
            plot(data, xlabel, ylabel=ylabel, ignore=ignore, xlim=xlim.get(xlabel), ylim=ylim,
                 suffix=suffix)

plot_f_map('ml1m', ylim_f=1.0, xlim={'time': 2000})

plot_f_map('ml10m', ylim_f=3.0)

plot_f_map('wiki-simple', ylim_f=4.5)



