from sklearn import datasets, preprocessing
import pdb
import numpy as np
import math
import scipy.io as sio
import os
import math
from scipy.sparse import csr_matrix
import tensorflow.compat.v1 as tf
from tensorflow.compat.v1 import SparseTensorValue as STV
import xclib.evaluation.xc_metrics as xc_metrics

MASK_MIN = -np.inf

class Y(object):
    def __init__(self, m, n, nnzs, indices=None, indptrs=None, nonzero=None):
        self.m = m
        self.n = n
        self.nnzs = nnzs

        if indices and indptrs and nonzero:
            nonzero = np.array(nonzero)

            self.Label = csr_matrix((np.ones(nnzs, dtype=float), nonzero), shape=(m, n))
            self.Sigma = csr_matrix((np.ones(nnzs, dtype=float), nonzero), shape=(m, n))

            self.Y_hat = csr_matrix((np.ones(nnzs, dtype=float), nonzero), shape=(m, n))
            self.Y_imp = csr_matrix((np.ones(nnzs, dtype=float), nonzero), shape=(m, n))



class ConfigClass(object):
    def __init__(self, args, m, n, nnzs, skip_mask):
        super(ConfigClass, self).__init__()
        self.args = args
        self.iter_max = args.iter_max
        self.skip_mask = skip_mask

        self.seed = args.seed

        if self.seed is None:
            print('You choose not to specify a random seed.'+\
                'A different result is produced after each run.')
        elif isinstance(self.seed, int) and self.seed >= 0:
            print('You specify random seed {}.'.format(self.seed))
        else:
            raise ValueError('Only accept None type or nonnegative integers for'+\
                    ' random seed argument!')

        self.u_set = args.u_set
        self.v_set = args.v_set
        self.val_set = args.val_set

        self.m = int(m)
        self.n = int(n)

        self.nnzs = int(nnzs)

        self.r = args.r

        self.GNsize = min(args.GNsize, self.m, self.n)
        self.C = args.C
        self.net_u = args.net_u
        self.net_v = args.net_v

        self.k = args.k
        self.k1 = args.k1

        self.xi = 0.1
        self.CGmax = args.CGmax
        self.omega=args.omega
        self._lambda = args._lambda
        self.drop = args.drop
        self.boost = args.boost
        self.eta = args.eta
        self.lr = args.lr
        self.alpha = args.alpha
        self.from_prev = args.from_prev
        self.thres = args.thres
        self.lr_decay = args.lr_decay
        self.b_ratio = args.b_ratio

        self.bsize = args.bsize

        rho = self.b_ratio ** 0.5
        self.sogram_bsize = round(rho * self.nnzs)

        self.bsize_i = round(self.m*(self.b_ratio**0.5))
        self.bsize_j = round(self.n*(self.b_ratio**0.5))
        #self.bsize_j = self.n

        self.gramian_alpha = 1 if self.sogram_bsize == self.nnzs else args.gramian_alpha
        if args.momentum < 0:
            raise ValueError('Momentum needs to be larger than 0!')
        self.momentum = args.momentum

        self.loss = args.loss
        if self.loss not in ('MSELoss', 'CrossEntropy'):
            raise ValueError('Unrecognized loss type!')
        self.optim = args.optim
        if self.optim not in ('SGD', 'NewtonCG', 'Adam', 'Sogram', 'Minibatch'):
            raise ValueError('Only support SGD, Adam, NewtonCG, Sogram, and Minibatch optimizer!')

        self.log_file = args.log_file
        self.model_file = args.model_file
        self.screen_log_only = args.screen_log_only
        self.profile = args.profile
        self.adagrad = args.adagrad
        if self.adagrad == -1:
            if self.optim == 'NewtonCG':
                self.adagrad = 0
            else:
                self.adagrad = 1

        if self.screen_log_only:
            print('You choose not to store running log. Only store model to {}'.format(self.log_file))
        else:
            print('Saving log to: {}'.format(self.log_file))
            dir_name, _ = os.path.split(self.log_file)
            if dir_name and not os.path.isdir(dir_name):
                os.makedirs(dir_name, exist_ok=True)

        dir_name, _ = os.path.split(self.model_file)
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name, exist_ok=True)
        self.elapsed_time = 0.0

def gen_label_matrix(labels_list, max_label_idx):
    m  = len(labels_list)
    n = -1

    indices = []
    indptrs = [1]
    rows = []
    columns = []

    nnz = 0
    row_idx = 0

    for labels in labels_list:
        for label in labels:
            rows.append(int(row_idx))
            columns.append(int(label))
        this_nnzs = len(labels)
        nnz += this_nnzs
        indptrs.append(indptrs[-1]+this_nnzs)
        indices += list(labels)
        n = max(n, max(labels))
        row_idx += 1
    return Y(int(m), max(int(n+1), max_label_idx), nnz, indices, indptrs, [rows, columns])

def read_data(filename, has_label=True, max_feat_idx = -1, max_label_idx = -1):
    """
    args:
        filename: the path where .mat files are stored
    """
    X, labels = datasets.load_svmlight_file(filename, zero_based=True, multilabel=True)
    X = csr_matrix((X.data, X.nonzero()), shape=(X.shape[0], max(X.shape[1], max_feat_idx)))
    X = preprocessing.normalize(X, norm="l2")
    if has_label:
        return X, gen_label_matrix(labels, max_label_idx)
    else:
        return X

def _cal_output(sess, input, input_placeholder, output, bsize):
    num_data = input.shape[0]
    num_segment = math.ceil(num_data/bsize)
    output_mat = np.zeros((num_data, output.shape[1].value), dtype=np.float64)
    for i in range(num_segment):
        idx = np.arange(i * bsize, min((i+1) * bsize, num_data))
        start,end = idx[0],idx[-1]+1
        batch_input = input[start:end]
        batch_input = STV(np.array(batch_input.nonzero()).transpose(), batch_input.data, batch_input.shape)
        output_mat[start:end] = sess.run(output, feed_dict={input_placeholder:batch_input})
    return output_mat

def gen_skip_mask(Y, skip_mask=None):
    if skip_mask == None:
        return 
    skip_mask_file = open(skip_mask, 'r')
    lines = skip_mask_file.readlines()
    mask = []
    for l in lines:
        mask.append(int(l))
    mask = np.array(mask)
    return Y[mask]

def predict(sess, network, Yte, Ute, V, config, k=5):
    u,v,p,q = network
    bsize, skip_mask = 2**17, config.skip_mask

    P = _cal_output(sess, Ute, u, p, bsize)
    Q = _cal_output(sess, V, v, q, bsize)
    
    # score_mat = P.dot(Q.T)
    m, n = P.shape[0], Q.shape[0]
    segment_m, segment_n = math.ceil(m/bsize), math.ceil(n/bsize)
    precision, ndcg = np.zeros(k,dtype=P.dtype), np.zeros(k,dtype=P.dtype)

    for i in range(segment_m):
        i_start, i_end = i*bsize, min((i+1)*bsize, m)
        score_mat = np.zeros((i_end-i_start, n), dtype=P.dtype)
        for j in range(segment_n):
            j_start, j_end = j*bsize, min((j+1)*bsize, n)
            score_mat[:, j_start:j_end] = P[i_start:i_end].dot(Q[j_start:j_end].T)
        if skip_mask is not None:
            score_mat[skip_mask[i_start:i_end].nonzero()] = MASK_MIN
        _precision = xc_metrics.precision(score_mat, Yte[i_start:i_end], k=k)
        _ndcg = xc_metrics.ndcg(score_mat, Yte[i_start:i_end], k=k)
        precision += (i_end-i_start)*_precision
        ndcg += (i_end-i_start)*_ndcg
    precision /= m
    ndcg /= m
    return precision, ndcg

def vectorize(tensors):
    if isinstance(tensors, list) or isinstance(tensors, tuple):
        vector = [tf.reshape(tensor, [-1]) for tensor in tensors]
        return tf.concat(vector, 0)
    else:
        return tensors

def inverse_vectorize(vector, param):
    if isinstance(vector, list):
        return vector
    else:
        tensors = []
        offset = 0
        num_total_param = np.sum([np.prod(p.shape.as_list()) for p in param])
        for p in param:
            numel = np.prod(p.shape.as_list())
            tensors.append(tf.reshape(vector[offset: offset+numel], p.shape))
            offset += numel

        assert offset == num_total_param
        return tensors

def calc_norm(v):
    # default: frobenius norm
    if isinstance(v, list):
        norm = 0.
        for p in v:
            norm = norm + tf.norm(tensor=p)**2
        return norm**0.5
    else:
        return tf.norm(tensor=v)

def zero_tensor(g):
    return tf.assign(g, tf.zeros_like(g))
